﻿using System;
using System.Collections.Generic;
using System.Collections.Immutable;
using System.Diagnostics;
using System.IO;
using System.Linq;
using System.Text;
using LibGit2Sharp;

namespace git_fault
{
    public class FaultMeta
    {
        public string Status { get; set; }
        public string Title { get; set; }
        public string Priority { get; set; }
    }

    public class CommandHandlers : IDisposable
    {
        private readonly string _headRemoteName;
        private readonly string _tempPathToMerge;
        private readonly string _pathToLocalFaultRefs;
        private readonly string _pathToRemoteFaultRefs;

        private readonly Repository _repo;

        private readonly Func<bool, IRenderer> _renderer;

        private readonly string yamlHeader = @"---
Title: 
Priority: 1
Status: Open
---
";

        public CommandHandlers(string pathToRepo, IRenderer htmlrenderer, IRenderer textrenderer)
        {
            _repo = new Repository(pathToRepo);
            _repo.Config.Set("core.logAllRefUpdates", true);
            _tempPathToMerge = $"{_repo.Info.Path}/TempDirectoryForMerge";
            _pathToLocalFaultRefs = "refs/faults/local";
            _headRemoteName = _repo.Head.RemoteName;
            _pathToRemoteFaultRefs = $"refs/faults/{_repo.Head.RemoteName}";

            _renderer = asHtml => asHtml ? htmlrenderer : textrenderer;

            Directory.CreateDirectory(_tempPathToMerge);
        }

        public void Dispose()
        {
            _repo?.Dispose();
        }

        private void AskForTextInput()
        {
            var editor = _repo.Config.Get<string>("core.editor")?.Value.Replace("\"","").Replace("'",""); //TODO works on my machine
            if (string.IsNullOrEmpty(editor))
            {
               Console.WriteLine("No core.editor defined"); 
            }

            var parameter = ""; //TODO works on my machine
            Process.Start(editor, string.Concat(Path.Combine(_tempPathToMerge, "issue"), ' ', parameter))?.WaitForExit();

            while (!Validator.AllFieldsInYAMLFilled(File.ReadAllText(Path.Combine(_tempPathToMerge, "issue"))))
            {
                AskForTextInput();
            }
        }

        public void Fetch(bool as_html)
        {
            if (Directory.EnumerateFileSystemEntries(_tempPathToMerge).Any())
            {
                _renderer(as_html).Error("Merge Directory not empty");
            }

            _repo.Network.Fetch(_headRemoteName, new string[] {$"+{_pathToRemoteFaultRefs}/*:{_pathToRemoteFaultRefs}/*"});
        }

        public void Merge(bool as_html)
        {
            var myrefs = _repo.Refs.FromGlob($"{_pathToLocalFaultRefs}/*").ToImmutableList().Select(m =>
                new
                {
                    refIdent = m.CanonicalName.Split("/")[^1], shaTarget = m.ResolveToDirectReference().TargetIdentifier
                });

            var originrefs = _repo.Refs.FromGlob($"{_pathToRemoteFaultRefs}/*").ToImmutableList().Select(m =>
                new
                {
                    refIdent = m.CanonicalName.Split("/")[^1], shaTarget = m.ResolveToDirectReference().TargetIdentifier
                });

            var newRefs = originrefs.Where(originref => myrefs.All(myref => myref.refIdent != originref.refIdent));

            foreach (var reference in newRefs)
            {
                _repo.Refs.Add($"{_pathToLocalFaultRefs}/{reference.refIdent}", reference.shaTarget,
                    $"Merged new ref {reference.refIdent}", false);
            }

            var diffs = myrefs.AsQueryable()
                .Join(originrefs, originref => originref.refIdent, myref => myref.refIdent,
                    (o, t) => new {ours = o, theirs = t})
                .Where(pairs =>
                    pairs.ours.shaTarget != pairs.theirs.shaTarget);

            foreach (var diff in diffs)
            {
                var x = _repo.ObjectDatabase.MergeCommitsIntoIndex(_repo.Lookup<Commit>(diff.ours.shaTarget),
                    _repo.Lookup<Commit>(diff.theirs.shaTarget), new MergeTreeOptions());
                if (x.IsFullyMerged)
                {
                    WriteIndexToDisk(x, _tempPathToMerge);
                    var commit = CreateCommit("Merged", _tempPathToMerge, new[] {_repo.Lookup<Commit>(diff.ours.shaTarget), _repo.Lookup<Commit>(diff.theirs.shaTarget)});
                    _repo.Refs.UpdateTarget(_repo.Refs[$"{_pathToLocalFaultRefs}/{diff.ours.refIdent}"], commit.Id.Sha, "Merged");
                    Directory.Delete(_tempPathToMerge, true);
                    Directory.CreateDirectory(_tempPathToMerge);
                }
                else
                {
                    foreach (var conflict in x.Conflicts)
                    {
                        if (_repo.Lookup<Commit>(diff.ours.shaTarget)[conflict.Ours.Path].Target.Peel<Blob>().IsBinary ||
                            _repo.Lookup<Commit>(diff.theirs.shaTarget)[conflict.Theirs.Path].Target.Peel<Blob>()
                                .IsBinary)
                        {
                            var b1 = _repo.Lookup<Commit>(diff.ours.shaTarget)[conflict.Ours.Path].Target.Peel<Blob>();
                            var b2 = _repo.Lookup<Commit>(diff.theirs.shaTarget)[conflict.Theirs.Path].Target.Peel<Blob>();

                            using var b1F = File.Create($"{_tempPathToMerge}/ours_{conflict.Ours.Path}");
                            b1.GetContentStream().CopyTo(b1F);
                            using var b2F = File.Create($"{_tempPathToMerge}/theirs_{conflict.Theirs.Path}");
                            b2.GetContentStream().CopyTo(b2F);
                        }
                        else
                        {
                            var c1 = _repo.Lookup<Commit>(diff.ours.shaTarget);
                            var c2 = _repo.Lookup<Commit>(diff.theirs.shaTarget);
                            var p = _repo.Diff.Compare(c1[conflict.Ours.Path].Target.Peel<Blob>(), c2[conflict.Theirs.Path].Target.Peel<Blob>(), new CompareOptions() {IncludeUnmodified = true, ContextLines = int.MaxValue});
                            File.WriteAllText($"{_tempPathToMerge}/{conflict.Ours.Path}", p.Patch);
                        }

                        AskForTextInput();
                        
                        UpdateFaultMerge(diff.theirs.refIdent, _tempPathToMerge, new Commit[]{_repo.Lookup<Commit>(diff.ours.shaTarget), _repo.Lookup<Commit>(diff.theirs.shaTarget)});
                        UpdateFault(diff.theirs.refIdent, null, _tempPathToMerge, as_html);
                        Directory.Delete(_tempPathToMerge, true);
                        Directory.CreateDirectory(_tempPathToMerge);
                    }
                }
            }
        }

        private void WriteIndexToDisk(TransientIndex transientIndex, string tempPathToMerge)
        {
            foreach (var entry in transientIndex)
            {
                switch (entry.Mode)
                {
                    case Mode.Directory:
                    {
                        Directory.CreateDirectory(Path.Combine(tempPathToMerge, entry.Path));
                        break;
                    }
                    case Mode.Nonexistent:
                    {
                        break;
                    }
                    case Mode.GitLink:
                    {
                        break;
                    }
                    case Mode.SymbolicLink:
                        break;
                    default:
                    {
                        var b = _repo.Lookup<Blob>(entry.Id);
                        Directory.CreateDirectory(Path.Combine(tempPathToMerge, Path.GetDirectoryName(entry.Path)));
                        var f = File.Create(Path.Combine(tempPathToMerge, entry.Path));
                        b.GetContentStream().CopyTo(f);
                        f.Dispose();
                        break;
                    }
                }
            }
        }

        public void Pull(bool as_html)
        {
            if (Directory.EnumerateFileSystemEntries(_tempPathToMerge).Any())
            {
                var error = $"Merge Directory ${_tempPathToMerge} not empty";
                _renderer(false).Error(error);
                throw new Exception(error);
            }

            Fetch(as_html);
            Merge(as_html);
        }

        public void Push()
        {
            if (Directory.EnumerateFileSystemEntries(_tempPathToMerge).Any())
            {
                return;
            }

            if (_repo.Head == null)
            {
                Console.WriteLine("Repo Head is null");
                return;
            }

            var remote = _repo.Network.Remotes[_headRemoteName];
            _repo.Network.Fetch(_headRemoteName, new string[] {$"{_pathToLocalFaultRefs}/*:{_pathToRemoteFaultRefs}/*"});

            IEnumerable<string> refs = _repo.Refs.FromGlob($"{_pathToLocalFaultRefs}/*").Join(_repo.Refs.FromGlob($"{_pathToLocalFaultRefs}/*"), outer => Path.GetFileName(outer.CanonicalName), inner =>  Path.GetFileName(inner.CanonicalName), (outer, inner) => outer + ":" + _pathToRemoteFaultRefs + "/"+ Path.GetFileName(inner.CanonicalName));
            _repo.Network.Push(remote, refs);
        }

        private Commit CreateCommit(string msg, string path, IEnumerable<Commit> parents)
        {
            using var textStream = new MemoryStream();

            var tree = CreateTreeFromPath(path, _repo);
            Signature author = _repo.Config.BuildSignature(DateTimeOffset.Now);
            var commit = _repo.ObjectDatabase.CreateCommit(author, author, msg, tree, parents, true);

            return commit;
        }

        private Tree CreateTreeFromPath(string path, Repository repo)
        {
            var files = Directory.GetFiles(path);
            var directories = Directory.GetDirectories(path);
            var td = new TreeDefinition();
            foreach (var file in files)
            {
                td.Add(Path.GetFileName(file), file, Mode.NonExecutableFile);
            }

            foreach (var directory in directories)
            {
                var st = CreateTreeFromPath(directory, repo);
                td.Add(Path.GetDirectoryName(directory), st);
            }

            return repo.ObjectDatabase.CreateTree(td);
        }

        public void CreateFault(string text, string path, bool as_html)
        {
            if (!string.IsNullOrEmpty(path) && !string.IsNullOrEmpty(text))
            {
                _renderer(false).Error("Only Text OR Path allowed");
            }

            if (string.IsNullOrEmpty(path) && string.IsNullOrEmpty(text))
            {
                _renderer(false).Error("Text OR Path needed");
            }

            if (!string.IsNullOrEmpty(text))
            {
                CreateFaultTextOnly(text, as_html);
            }

            if (!string.IsNullOrEmpty(path))
            {
                CreateFaultPathOnly(Path.GetFullPath(path), as_html);
            }
        }

        private void CreateFaultPathOnly(string path, bool as_html)
        {
            if (!Validator.AllFieldsInYAMLFilled(File.ReadAllText(Path.Combine(path, "issue"))))
            {
                AskForTextInput();
            }

            var msg = "created fault";
            var commit = CreateCommit(msg, path, new Commit[] { });
            _repo.Refs.Add($"{_pathToLocalFaultRefs}/{commit.Sha}", commit.Id.Sha, msg, false);

            Directory.GetFiles(path).Where(f => f != "issue").ToImmutableList().ForEach(f => CreateAppendix(commit.Sha, f, as_html));

            _renderer(as_html).NewFaultCreated(commit.Sha, File.ReadAllText(Path.Combine(path, "issue")));
        }

        private void CreateFaultTextOnly(string text, bool as_html)
        {
            File.WriteAllText(Path.Combine(_tempPathToMerge, "issue"), text, Encoding.UTF8);
            if (!Validator.AllFieldsInYAMLFilled(File.ReadAllText(Path.Combine(_tempPathToMerge, "issue"))))
            {
                File.WriteAllText(Path.Combine(_tempPathToMerge, "issue"), yamlHeader, Encoding.UTF8);
                File.AppendAllText(Path.Combine(_tempPathToMerge, "issue"), text, Encoding.UTF8);
                AskForTextInput();
            }

            var msg = "created fault";
            var commit = CreateCommit(msg, _tempPathToMerge, new Commit[] { });
            _repo.Refs.Add($"{_pathToLocalFaultRefs}/{commit.Sha}", commit.Id.Sha, msg, false);

            _renderer(as_html).NewFaultCreated(commit.Sha, text);

            ProgramOpener.DeleteDirectory(_tempPathToMerge, true);
        }

        private (Reference reference, Commit commit) GetRefAndCommitFromId(string id)
        {
            var reference1 = _repo.Refs[$"{_pathToLocalFaultRefs}/{id}"];
            var reference2 = _repo.Refs[$"{_pathToRemoteFaultRefs}/{id}"];
            var parent1 = reference1?.ResolveToDirectReference().Target.Peel<Commit>();
            var parent2 = reference2?.ResolveToDirectReference().Target.Peel<Commit>();

            return reference1 == null ? (reference2, parent2) : (reference1, parent1);
        }

        private void UpdateFaultMerge(string id, string path, Commit[] parent)
        {
            var tree = CreateTreeFromPath(path, _repo);

            Signature author = _repo.Config.BuildSignature(DateTimeOffset.Now);
            var msg = "Updated Issue";
            var commit = _repo.ObjectDatabase.CreateCommit(author, author, msg, tree, new[] {parent[0], parent[1]}, true);
            var refs = _repo.Refs[$"{_pathToLocalFaultRefs}/{id}"];
            _repo.Refs.UpdateTarget(refs.CanonicalName, commit.Id.Sha, msg);
        }
        
        public void UpdateFault(string id, string newText, string path, bool as_html)
        {
            if (string.IsNullOrEmpty(path))
            {
                var refCommit = GetRefAndCommitFromId(id);

                if (!string.IsNullOrEmpty(newText))
                {
                    File.WriteAllText(Path.Combine(_tempPathToMerge, "issue"), newText, Encoding.UTF8);
                }
                else
                {
                    File.WriteAllText(Path.Combine(_tempPathToMerge, "issue"), refCommit.commit.Tree["issue"]?.Target.Peel<Blob>().GetContentText(),
                        Encoding.UTF8);
                    AskForTextInput();
                }

                if (!Validator.AllFieldsInYAMLFilled(File.ReadAllText(Path.Combine(_tempPathToMerge, "issue"))))
                {
                    AskForTextInput();
                }

                var msg = "Updated Issue";
                var commit = CreateCommit(msg, _tempPathToMerge, new[] {refCommit.commit});
                _repo.Refs.UpdateTarget(refCommit.reference, commit.Id.Sha, msg);

                var oldText = refCommit.commit.Tree["issue"]?.Target.Peel<Blob>().GetContentText();

                Directory.Delete(_tempPathToMerge, true);
                Directory.CreateDirectory(_tempPathToMerge);

                _renderer(as_html).FaultUpdated(id, oldText, newText);
            }
            else if (!string.IsNullOrEmpty(path))
            {
                var reference = _repo.Refs[$"{_pathToLocalFaultRefs}/{id}"];
                var parent = reference.ResolveToDirectReference().Target.Peel<Commit>();

                if (!Validator.AllFieldsInYAMLFilled(File.ReadAllText(Path.Combine(path, "issue"))))
                {
                    AskForTextInput();
                }

                var tree = CreateTreeFromPath(path, _repo);

                Signature author = _repo.Config.BuildSignature(DateTimeOffset.Now);
                var msg = "Updated Issue";
                var commit = _repo.ObjectDatabase.CreateCommit(author, author, msg, tree, new[] {parent}, true);
                _repo.Refs.UpdateTarget(reference, commit.Id.Sha, msg);

                var oldText = parent.Tree["issue"]?.Target.Peel<Blob>().GetContentText();
                _renderer(as_html).FaultUpdated(id, oldText, newText);
            }
        }

        public void DeleteFault(string id, bool as_html)
        {
            if (Directory.EnumerateFileSystemEntries(_tempPathToMerge).Any())
            {
                _renderer(as_html).Error("Temp path to merge not empty");
            }

            var treeDefinition = new TreeDefinition();
            var tree = _repo.ObjectDatabase.CreateTree(treeDefinition);
            Signature author = _repo.Config.BuildSignature(DateTimeOffset.Now);

            var refCommit = GetRefAndCommitFromId(id);
            var child = _repo.ObjectDatabase.CreateCommit(author, author, $"Removed issue {refCommit.commit.Sha}", tree, new List<Commit> {refCommit.commit}, true);
            _repo.Refs.UpdateTarget(refCommit.reference, child.Id, "Removed Issue");

            var oldText = refCommit.commit.Tree["issue"]?.Target.Peel<Blob>().GetContentText();
            _renderer(as_html).FaultDeleted(id, oldText);
        }

        public void LsFault(bool full, bool as_html)
        {
            var refs = _repo.Refs.FromGlob($"{_pathToLocalFaultRefs}/*");

            var faults = new List<(string, string, List<(string, string)>)>();
            foreach (var refe in refs)
            {
                var appendices = new List<(string, string)>();
                var fault = refe.ResolveToDirectReference().Target.Peel<Commit>().Tree["issue"].Target.Peel<Blob>()
                    .GetContentText();
                var appendix = refe.ResolveToDirectReference().Target.Peel<Commit>().Tree["appendix"]?.Target
                    .Peel<Tree>();
                if (appendix != null)
                {
                    foreach (var item in appendix)
                    {
                        appendices.Add((item.Target.Peel<Blob>().Id.Sha, item.Name));
                    }
                }

                faults.Add((refe.CanonicalName, fault, appendices));
            }

            _renderer(as_html).FaultsListed(full, faults);
        }

        public void ShowFault(string id, bool as_html)
        {
            var refCommit = GetRefAndCommitFromId(id);
            var issue = refCommit.commit.Tree["issue"]?.Target.Peel<Blob>();
            _renderer(as_html).ShowFault(issue.GetContentText());
        }

        public void CreateAppendix(string id, string file, bool as_html)
        {
            var blob = _repo.ObjectDatabase.CreateBlob(file);

            var refCommit = GetRefAndCommitFromId(id);
            var parent = refCommit.reference.ResolveToDirectReference().Target.Peel<Commit>();
            var oldAppendixTD = parent.Tree["appendix"];
            var newTD = TreeDefinition.From(parent.Tree);
            var msg = "New Appendix created";

            var newAppendixTD = oldAppendixTD == null
                ? new TreeDefinition()
                : TreeDefinition.From(oldAppendixTD.Target.Peel<Tree>());
            newAppendixTD.Add(Path.GetFileName(file), blob, Mode.NonExecutableFile);
            var appendixT = _repo.ObjectDatabase.CreateTree(newAppendixTD);
            newTD.Add("appendix", appendixT);

            var newT = _repo.ObjectDatabase.CreateTree(newTD);

            Signature author = _repo.Config.BuildSignature(DateTimeOffset.Now);

            var newcommit =
                _repo.ObjectDatabase.CreateCommit(author, author, msg, newT, new List<Commit> {parent}, true);
            _repo.Refs.UpdateTarget(refCommit.reference, newcommit.Id.Sha, msg);
        }

        public void DeleteAppendix(string id, string file, bool as_html)
        {
            if (Directory.EnumerateFileSystemEntries(_tempPathToMerge).Any())
            {
                _renderer(as_html).Error("Merge Path not empty");
            }

            var refCommit = GetRefAndCommitFromId(id);
            var oldAppendixTD = refCommit.commit.Tree["appendix"];
            var newTD = TreeDefinition.From(refCommit.commit.Tree);
            var msg = "Appendix deleted";

            var newAppendixTD = TreeDefinition.From(oldAppendixTD.Target.Peel<Tree>());
            newAppendixTD.Remove(Path.GetFileName(file));
            var appendixT = _repo.ObjectDatabase.CreateTree(newAppendixTD);
            newTD.Add("appendix", appendixT);

            var newT = _repo.ObjectDatabase.CreateTree(newTD);

            Signature author = _repo.Config.BuildSignature(DateTimeOffset.Now);

            var newcommit = _repo.ObjectDatabase.CreateCommit(author, author, msg, newT, new List<Commit> {refCommit.commit}, true);
            _repo.Refs.UpdateTarget(refCommit.reference, newcommit.Id.Sha, msg);
        }

        public void ShowAppendix(string faultid, string name, bool as_html)
        {
            var refCommit = GetRefAndCommitFromId(faultid);

            var appendix = refCommit.commit.Tree["appendix"]?.Target.Peel<Tree>()[name]?.Target.Peel<Blob>();

            if (appendix == null)
            {
                return;
            }

            if (appendix.IsBinary)
            {
                var path = Path.Combine(System.IO.Path.GetTempPath(), name);
                var file = File.Create(path);
                appendix.GetContentStream().CopyTo(file);
                file.Close();
                ProgramOpener.Open(path);
                return;
            }

            _renderer(as_html).ShowAppendix(appendix.GetContentText(), name);
        }
    }
}